#!/bin/bash
########################################################################
# build_clang.sh
#
# Build clang and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_08_05.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: $*\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: $prog <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       $prog -h|--help [<product-dir>]

  Build $package and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking $prog.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking $prog.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "$@"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="$arg"; break
  done
  if [ -z "$product_dir" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> $product_dir\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "$product_dir/.upsfiles" ] && [ -w "$product_dir" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "$product_dir\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "$product_dir" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: $product_dir\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="$product_dir/ssibuildshims/$ssibuildshims_version"
  [ -f "$ssibuildshims_topdir/bin/ssi_functions" ] && \
    source "$ssibuildshims_topdir/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "$product_dir" && \
    setup ssibuildshims $ssibuildshims_version -z "$product_dir" || \
    ssi_die "unable to find ssibuildshims $ssibuildshims_version in
$product_dir and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "$product_dir" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  $trace verify_in_filesystem -x "$@" "$pkgdir/bin/clang-${pkgdotver%%.*}" \
    && $trace verify_in_filesystem -x "$@" "$pkgdir/bin/flang"
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "$maketar" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=(-d data:doc:include:scripts:example:share)
    $trace is_built "$@" && \
      $trace make_basic_distribution "$@" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Package-specific functions.
########################################################################

# ...

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
thisdir=$(cd "$(dirname "${BASH_SOURCE[0]}")" && pwd -P)
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
. $thisdir/version_info.sh || \
  ssi_die "unable to load version info from $thisdir/version_info"
####################################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "$@"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_(CONFIG|BUILD|INSTALL)

    Verbose output from the corresponding step.

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.
EOF
##################

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
$trace define_basics NO_QUAL NO_DEBUG_SOURCE TRACK_PROGRESS || exit # *C*
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
$trace fake_declare_product "${ssi_args[@]}" -x -- || exit
####################################

####################################
# Build-only dependencies. *C*
$trace setup_cmake --latest={--ge,3.15} || exit
$trace setup_ninja --latest && (( want_ninja = 1 )) || \
  ssi_die "Ninja required for JOB_POOLS"

(( WANT_TESTS )) && $trace setup_python --latest
[ "$OS1" = Darwin ] || \
  ! ver_greater 5.1.0 $(gcc -dumpversion) || \
  $trace setup_gcc --latest={--ge,6.1.0,--le,12.1.0} || \
  ssi_die "unable to find a suitable compiler to build LLVM/CLang $pkgdotver"
####################################

# Print active products.
echo
ups active
echo

# Possible short-circuit if we're already built and declared.
$trace maybe_make_distribution --silent

ssi_info "building $package for $OS-$plat-$qualdir (flavor $pkgflvr)"

# Ensure directories: pkgdir, sourcedir, blddir.
$trace init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

$trace cd "$sourcedir" \
  || ssi_die "unable to cd to source directory $sourcedir"

# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.

eval patches=($(find "$patchdir" -name '*.patch' -printf "'%P'\n"))
$trace prepare_source ${ssi_args[*]:+"${ssi_args[@]}"} -P "$patchdir" -- \
  "$tardir/$pkgtarfile" \
  ${patches[*]:+"${patches[@]}"} \
  || exit

# Bail early if DEBUG_PATCHES is set.
[ -n "$DEBUG_PATCHES" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.

##################
# Configure parallelism. *C*
ncores=$(ncores)

# For cmake --build:
[[ -n "$CMAKE_BUILD_PARALLEL_LEVEL" ]] || \
  export CMAKE_BUILD_PARALLEL_LEVEL=$ncores

# For ctest:
[[ -n "$CTEST_PARALLEL_LEVEL" ]] || \
  export CTEST_PARALLEL_LEVEL=$ncores
##################

##################
# Basic CMake steering options.
cmake_options=()
(( want_ninja )) && cmake_options+=(-GNinja)
##################

##################
# General compilation, linking options, etc.
prefix="$pkgdir"

if [ "$OS1" = Darwin ]; then
  cc=/usr/bin/clang
  cxx=/usr/bin/clang++
  DY=DY
else
  cc=gcc
  cxx=g++
  DY=
  ver_greater 9.0.0 $(gcc -dumpversion) || \
    cmake_options+=(-DLLVM_ENABLE_WARNINGS:BOOL=OFF)
  target_extra=
fi

extra_cmake_args=(-Wno-dev -DCMAKE_EXPORT_COMPILE_COMMANDS:BOOL=ON)
if (( DEBUG_CONFIG )); then
  extra_cmake_args+=(--trace)
  if (( DEBUG_CONFIG > 1 )); then
    log_level=DEBUG
    extra_cmake_args+=(--trace-expand)
  else
    log_level=VERBOSE
  fi
   extra_cmake_args+=(--log-level=$log_level)
fi

ssi_info "getting host triple"
# Shorthand.
src_top="$sourcedir/$pkgtarname/llvm"
host_triple="$(cmake -DCMAKE_MODULE_PATH="$src_top/cmake/modules" -DLLVM_MAIN_SRC_DIR="$src_top" -P "$patchdir/get_triple.cmake" | sed -nEe 's&^-- host_triple=&&p')"
[ -n "$host_triple" ] || ssi_die "unable to ascertain host triple"
ssi_info "host triple: $host_triple"

linker_flags=(
  "-L$blddir/lib"
  "-L$blddir/lib/$host_triple"
)

# Options for stage 1 (bootstrap).
cmake_options+=(
  ${extra_cmake_args[*]:+"${extra_cmake_args[@]}"}
  -DCMAKE_INSTALL_PREFIX="$pkgdir"
  -DCMAKE_BUILD_RPATH="\$ORIGIN/../lib;\$ORIGIN/../lib/$host_triple"
  -DCMAKE_BUILD_RPATH_USE_ORIGIN:BOOL=ON
  -DCMAKE_CXX_COMPILER=$cxx
  -DCMAKE_CXX_STANDARD=17
  -DCMAKE_C_COMPILER=$cc
)

linker_flags=(
  "-L$blddir/tools/clang/stage2-bins/lib"
  "-L$blddir/tools/clang/stage2-bins/lib/$host_triple"
  ${linker_flags[*]:+"${linker_flags[@]}"}
)
bootstrap_cmake_args=(
)
[ -n "${linker_flags[*]}" ] && \
  bootstrap_cmake_args+=(
    -DCMAKE_EXE_LINKER_FLAGS="${linker_flags[*]}"
    -DCMAKE_MODULE_LINKER_FLAGS="${linker_flags[*]}"
    -DCMAKE_SHARED_LINKER_FLAGS="${linker_flags[*]}"
  )

# Must come last.
bootstrap_cmake_args+=(-C "$patchdir/FNAL-stage2.cmake")

OIFS="$IFS"
IFS=";"
bootstrap_cmake_args_s=${bootstrap_cmake_args[*]:+"${bootstrap_cmake_args[*]}"}
IFS="$OIFS"

# Options for stage 2.
cmake_options+=(
  -DBOOTSTRAP_CMAKE_BUILD_RPATH="\$ORIGIN/../lib;\$ORIGIN/../lib/$host_triple;\$ORIGIN/../../../../lib;\$ORIGIN/../../../../lib/$host_triple"
  -DBOOTSTRAP_CMAKE_BUILD_RPATH_USE_ORIGIN:BOOL=ON
  -DBOOTSTRAP_CMAKE_CXX_STANDARD=17
  -DBOOTSTRAP_CMAKE_INSTALL_PREFIX="$pkgdir"
  -DBOOTSTRAP_LLVM_LDFLAGS="-Wl,--no-keep-memory"
  ${bootstrap_cmake_args_s:+-DCLANG_BOOTSTRAP_CMAKE_ARGS="$bootstrap_cmake_args_s"}
)
##################
####################################

####################################
# Configure, build, and install.
#
# Use $blddir/$pkgtarname for in-source builds, $blddir otherwise.

$trace cd "$blddir" || \
  ssi_die "unable to cd to build directory $blddir"

src_top="$sourcedir/$pkgtarname/llvm"

##################
# Configure. *C*

# N.B. It is *VERY IMPORTANT* that all -D args come before the -C, so
#      add the -C here so we can never put any in late accidentally.
$trace cmake "${cmake_options[@]}" \
         -C "$patchdir/FNAL.cmake" \
         "$src_top" || \
  ssi_die "configure failed"
##################

##################
# Build. *C*
if (( DEBUG_BUILD )); then
  export VERBOSE=1
else
  unset VERBOSE
fi

$trace cmake --build "$blddir" -t stage2$target_extra \
  || ssi_die "build failed"
##################

##################
# Install. *C*
if (( DEBUG_INSTALL )); then
  export VERBOSE=1
else
  unset VERBOSE
fi

$trace cmake --build "$blddir" -t stage2-install$target_extra \
  || ssi_die "install failed"
##################

ssi_info "finished building $package $pkgver"

####################################
# Pre-declaration processing. *C*
#
# Move things, install from contrib, clean build products from in-source
# builds, etc.
####################################

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
$trace declare_product -x -- || exit
##################

####################################
# Post-declaration processing. *C*
#
# Operations requiring the *final* declaration and setup (as opposed to
# the pre-build fake declaration and setup). This should be rare.
####################################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
$trace is_built && $trace ssi_build_completed && \
  ssi_info "$package is installed at $pkgdir" && \
  $trace maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
